@@ -30,29 +30,337 @@ import functools
 import os
 import typing
 
-# Low level API
-from ._pygit2 import *
-from ._pygit2 import _cache_enums
-
 # High level API
 from . import enums
 from ._build import __version__
+
+# Low level API
+from ._pygit2 import (
+    GIT_APPLY_LOCATION_BOTH,
+    GIT_APPLY_LOCATION_INDEX,
+    GIT_APPLY_LOCATION_WORKDIR,
+    GIT_BLAME_FIRST_PARENT,
+    GIT_BLAME_IGNORE_WHITESPACE,
+    GIT_BLAME_NORMAL,
+    GIT_BLAME_TRACK_COPIES_ANY_COMMIT_COPIES,
+    GIT_BLAME_TRACK_COPIES_SAME_COMMIT_COPIES,
+    GIT_BLAME_TRACK_COPIES_SAME_COMMIT_MOVES,
+    GIT_BLAME_TRACK_COPIES_SAME_FILE,
+    GIT_BLAME_USE_MAILMAP,
+    GIT_BLOB_FILTER_ATTRIBUTES_FROM_COMMIT,
+    GIT_BLOB_FILTER_ATTRIBUTES_FROM_HEAD,
+    GIT_BLOB_FILTER_CHECK_FOR_BINARY,
+    GIT_BLOB_FILTER_NO_SYSTEM_ATTRIBUTES,
+    GIT_BRANCH_ALL,
+    GIT_BRANCH_LOCAL,
+    GIT_BRANCH_REMOTE,
+    GIT_CHECKOUT_ALLOW_CONFLICTS,
+    GIT_CHECKOUT_CONFLICT_STYLE_DIFF3,
+    GIT_CHECKOUT_CONFLICT_STYLE_MERGE,
+    GIT_CHECKOUT_CONFLICT_STYLE_ZDIFF3,
+    GIT_CHECKOUT_DISABLE_PATHSPEC_MATCH,
+    GIT_CHECKOUT_DONT_OVERWRITE_IGNORED,
+    GIT_CHECKOUT_DONT_REMOVE_EXISTING,
+    GIT_CHECKOUT_DONT_UPDATE_INDEX,
+    GIT_CHECKOUT_DONT_WRITE_INDEX,
+    GIT_CHECKOUT_DRY_RUN,
+    GIT_CHECKOUT_FORCE,
+    GIT_CHECKOUT_NO_REFRESH,
+    GIT_CHECKOUT_NONE,
+    GIT_CHECKOUT_RECREATE_MISSING,
+    GIT_CHECKOUT_REMOVE_IGNORED,
+    GIT_CHECKOUT_REMOVE_UNTRACKED,
+    GIT_CHECKOUT_SAFE,
+    GIT_CHECKOUT_SKIP_LOCKED_DIRECTORIES,
+    GIT_CHECKOUT_SKIP_UNMERGED,
+    GIT_CHECKOUT_UPDATE_ONLY,
+    GIT_CHECKOUT_USE_OURS,
+    GIT_CHECKOUT_USE_THEIRS,
+    GIT_CONFIG_HIGHEST_LEVEL,
+    GIT_CONFIG_LEVEL_APP,
+    GIT_CONFIG_LEVEL_GLOBAL,
+    GIT_CONFIG_LEVEL_LOCAL,
+    GIT_CONFIG_LEVEL_PROGRAMDATA,
+    GIT_CONFIG_LEVEL_SYSTEM,
+    GIT_CONFIG_LEVEL_WORKTREE,
+    GIT_CONFIG_LEVEL_XDG,
+    GIT_DELTA_ADDED,
+    GIT_DELTA_CONFLICTED,
+    GIT_DELTA_COPIED,
+    GIT_DELTA_DELETED,
+    GIT_DELTA_IGNORED,
+    GIT_DELTA_MODIFIED,
+    GIT_DELTA_RENAMED,
+    GIT_DELTA_TYPECHANGE,
+    GIT_DELTA_UNMODIFIED,
+    GIT_DELTA_UNREADABLE,
+    GIT_DELTA_UNTRACKED,
+    GIT_DESCRIBE_ALL,
+    GIT_DESCRIBE_DEFAULT,
+    GIT_DESCRIBE_TAGS,
+    GIT_DIFF_BREAK_REWRITES,
+    GIT_DIFF_BREAK_REWRITES_FOR_RENAMES_ONLY,
+    GIT_DIFF_DISABLE_PATHSPEC_MATCH,
+    GIT_DIFF_ENABLE_FAST_UNTRACKED_DIRS,
+    GIT_DIFF_FIND_ALL,
+    GIT_DIFF_FIND_AND_BREAK_REWRITES,
+    GIT_DIFF_FIND_BY_CONFIG,
+    GIT_DIFF_FIND_COPIES,
+    GIT_DIFF_FIND_COPIES_FROM_UNMODIFIED,
+    GIT_DIFF_FIND_DONT_IGNORE_WHITESPACE,
+    GIT_DIFF_FIND_EXACT_MATCH_ONLY,
+    GIT_DIFF_FIND_FOR_UNTRACKED,
+    GIT_DIFF_FIND_IGNORE_LEADING_WHITESPACE,
+    GIT_DIFF_FIND_IGNORE_WHITESPACE,
+    GIT_DIFF_FIND_REMOVE_UNMODIFIED,
+    GIT_DIFF_FIND_RENAMES,
+    GIT_DIFF_FIND_RENAMES_FROM_REWRITES,
+    GIT_DIFF_FIND_REWRITES,
+    GIT_DIFF_FLAG_BINARY,
+    GIT_DIFF_FLAG_EXISTS,
+    GIT_DIFF_FLAG_NOT_BINARY,
+    GIT_DIFF_FLAG_VALID_ID,
+    GIT_DIFF_FLAG_VALID_SIZE,
+    GIT_DIFF_FORCE_BINARY,
+    GIT_DIFF_FORCE_TEXT,
+    GIT_DIFF_IGNORE_BLANK_LINES,
+    GIT_DIFF_IGNORE_CASE,
+    GIT_DIFF_IGNORE_FILEMODE,
+    GIT_DIFF_IGNORE_SUBMODULES,
+    GIT_DIFF_IGNORE_WHITESPACE,
+    GIT_DIFF_IGNORE_WHITESPACE_CHANGE,
+    GIT_DIFF_IGNORE_WHITESPACE_EOL,
+    GIT_DIFF_INCLUDE_CASECHANGE,
+    GIT_DIFF_INCLUDE_IGNORED,
+    GIT_DIFF_INCLUDE_TYPECHANGE,
+    GIT_DIFF_INCLUDE_TYPECHANGE_TREES,
+    GIT_DIFF_INCLUDE_UNMODIFIED,
+    GIT_DIFF_INCLUDE_UNREADABLE,
+    GIT_DIFF_INCLUDE_UNREADABLE_AS_UNTRACKED,
+    GIT_DIFF_INCLUDE_UNTRACKED,
+    GIT_DIFF_INDENT_HEURISTIC,
+    GIT_DIFF_MINIMAL,
+    GIT_DIFF_NORMAL,
+    GIT_DIFF_PATIENCE,
+    GIT_DIFF_RECURSE_IGNORED_DIRS,
+    GIT_DIFF_RECURSE_UNTRACKED_DIRS,
+    GIT_DIFF_REVERSE,
+    GIT_DIFF_SHOW_BINARY,
+    GIT_DIFF_SHOW_UNMODIFIED,
+    GIT_DIFF_SHOW_UNTRACKED_CONTENT,
+    GIT_DIFF_SKIP_BINARY_CHECK,
+    GIT_DIFF_STATS_FULL,
+    GIT_DIFF_STATS_INCLUDE_SUMMARY,
+    GIT_DIFF_STATS_NONE,
+    GIT_DIFF_STATS_NUMBER,
+    GIT_DIFF_STATS_SHORT,
+    GIT_DIFF_UPDATE_INDEX,
+    GIT_FILEMODE_BLOB,
+    GIT_FILEMODE_BLOB_EXECUTABLE,
+    GIT_FILEMODE_COMMIT,
+    GIT_FILEMODE_LINK,
+    GIT_FILEMODE_TREE,
+    GIT_FILEMODE_UNREADABLE,
+    GIT_FILTER_ALLOW_UNSAFE,
+    GIT_FILTER_ATTRIBUTES_FROM_COMMIT,
+    GIT_FILTER_ATTRIBUTES_FROM_HEAD,
+    GIT_FILTER_CLEAN,
+    GIT_FILTER_DEFAULT,
+    GIT_FILTER_DRIVER_PRIORITY,
+    GIT_FILTER_NO_SYSTEM_ATTRIBUTES,
+    GIT_FILTER_SMUDGE,
+    GIT_FILTER_TO_ODB,
+    GIT_FILTER_TO_WORKTREE,
+    GIT_MERGE_ANALYSIS_FASTFORWARD,
+    GIT_MERGE_ANALYSIS_NONE,
+    GIT_MERGE_ANALYSIS_NORMAL,
+    GIT_MERGE_ANALYSIS_UNBORN,
+    GIT_MERGE_ANALYSIS_UP_TO_DATE,
+    GIT_MERGE_PREFERENCE_FASTFORWARD_ONLY,
+    GIT_MERGE_PREFERENCE_NO_FASTFORWARD,
+    GIT_MERGE_PREFERENCE_NONE,
+    GIT_OBJECT_ANY,
+    GIT_OBJECT_BLOB,
+    GIT_OBJECT_COMMIT,
+    GIT_OBJECT_INVALID,
+    GIT_OBJECT_OFS_DELTA,
+    GIT_OBJECT_REF_DELTA,
+    GIT_OBJECT_TAG,
+    GIT_OBJECT_TREE,
+    GIT_OID_HEX_ZERO,
+    GIT_OID_HEXSZ,
+    GIT_OID_MINPREFIXLEN,
+    GIT_OID_RAWSZ,
+    GIT_REFERENCES_ALL,
+    GIT_REFERENCES_BRANCHES,
+    GIT_REFERENCES_TAGS,
+    GIT_RESET_HARD,
+    GIT_RESET_MIXED,
+    GIT_RESET_SOFT,
+    GIT_REVSPEC_MERGE_BASE,
+    GIT_REVSPEC_RANGE,
+    GIT_REVSPEC_SINGLE,
+    GIT_SORT_NONE,
+    GIT_SORT_REVERSE,
+    GIT_SORT_TIME,
+    GIT_SORT_TOPOLOGICAL,
+    GIT_STASH_APPLY_DEFAULT,
+    GIT_STASH_APPLY_REINSTATE_INDEX,
+    GIT_STASH_DEFAULT,
+    GIT_STASH_INCLUDE_IGNORED,
+    GIT_STASH_INCLUDE_UNTRACKED,
+    GIT_STASH_KEEP_ALL,
+    GIT_STASH_KEEP_INDEX,
+    GIT_STATUS_CONFLICTED,
+    GIT_STATUS_CURRENT,
+    GIT_STATUS_IGNORED,
+    GIT_STATUS_INDEX_DELETED,
+    GIT_STATUS_INDEX_MODIFIED,
+    GIT_STATUS_INDEX_NEW,
+    GIT_STATUS_INDEX_RENAMED,
+    GIT_STATUS_INDEX_TYPECHANGE,
+    GIT_STATUS_WT_DELETED,
+    GIT_STATUS_WT_MODIFIED,
+    GIT_STATUS_WT_NEW,
+    GIT_STATUS_WT_RENAMED,
+    GIT_STATUS_WT_TYPECHANGE,
+    GIT_STATUS_WT_UNREADABLE,
+    GIT_SUBMODULE_IGNORE_ALL,
+    GIT_SUBMODULE_IGNORE_DIRTY,
+    GIT_SUBMODULE_IGNORE_NONE,
+    GIT_SUBMODULE_IGNORE_UNSPECIFIED,
+    GIT_SUBMODULE_IGNORE_UNTRACKED,
+    GIT_SUBMODULE_STATUS_IN_CONFIG,
+    GIT_SUBMODULE_STATUS_IN_HEAD,
+    GIT_SUBMODULE_STATUS_IN_INDEX,
+    GIT_SUBMODULE_STATUS_IN_WD,
+    GIT_SUBMODULE_STATUS_INDEX_ADDED,
+    GIT_SUBMODULE_STATUS_INDEX_DELETED,
+    GIT_SUBMODULE_STATUS_INDEX_MODIFIED,
+    GIT_SUBMODULE_STATUS_WD_ADDED,
+    GIT_SUBMODULE_STATUS_WD_DELETED,
+    GIT_SUBMODULE_STATUS_WD_INDEX_MODIFIED,
+    GIT_SUBMODULE_STATUS_WD_MODIFIED,
+    GIT_SUBMODULE_STATUS_WD_UNINITIALIZED,
+    GIT_SUBMODULE_STATUS_WD_UNTRACKED,
+    GIT_SUBMODULE_STATUS_WD_WD_MODIFIED,
+    LIBGIT2_VER_MAJOR,
+    LIBGIT2_VER_MINOR,
+    LIBGIT2_VER_REVISION,
+    LIBGIT2_VERSION,
+    AlreadyExistsError,
+    Blob,
+    Branch,
+    Commit,
+    Diff,
+    DiffDelta,
+    DiffFile,
+    DiffHunk,
+    DiffLine,
+    DiffStats,
+    FilterSource,
+    GitError,
+    InvalidSpecError,
+    Mailmap,
+    Note,
+    Object,
+    Odb,
+    OdbBackend,
+    OdbBackendLoose,
+    OdbBackendPack,
+    Oid,
+    Patch,
+    Refdb,
+    RefdbBackend,
+    RefdbFsBackend,
+    Reference,
+    RefLogEntry,
+    RevSpec,
+    Signature,
+    Stash,
+    Tag,
+    Tree,
+    TreeBuilder,
+    Walker,
+    Worktree,
+    _cache_enums,
+    discover_repository,
+    filter_register,
+    filter_unregister,
+    hash,
+    hashfile,
+    init_file_backend,
+    reference_is_valid_name,
+    tree_entry_cmp,
+)
 from .blame import Blame, BlameHunk
 from .blob import BlobIO
-from .callbacks import Payload, RemoteCallbacks, CheckoutCallbacks, StashApplyCallbacks
 from .callbacks import (
+    CheckoutCallbacks,
+    Payload,
+    RemoteCallbacks,
+    StashApplyCallbacks,
+    get_credentials,
     git_clone_options,
     git_fetch_options,
     git_proxy_options,
-    get_credentials,
 )
 from .config import Config
 from .credentials import *
-from .errors import check_error, Passthrough
-from .ffi import ffi, C
+from .errors import Passthrough, check_error
+from .ffi import C, ffi
 from .filter import Filter
 from .index import Index, IndexEntry
 from .legacyenums import *
+from .options import (
+    GIT_OPT_ADD_SSL_X509_CERT,
+    GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS,
+    GIT_OPT_ENABLE_CACHING,
+    GIT_OPT_ENABLE_FSYNC_GITDIR,
+    GIT_OPT_ENABLE_HTTP_EXPECT_CONTINUE,
+    GIT_OPT_ENABLE_OFS_DELTA,
+    GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION,
+    GIT_OPT_ENABLE_STRICT_OBJECT_CREATION,
+    GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION,
+    GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY,
+    GIT_OPT_GET_CACHED_MEMORY,
+    GIT_OPT_GET_EXTENSIONS,
+    GIT_OPT_GET_HOMEDIR,
+    GIT_OPT_GET_MWINDOW_FILE_LIMIT,
+    GIT_OPT_GET_MWINDOW_MAPPED_LIMIT,
+    GIT_OPT_GET_MWINDOW_SIZE,
+    GIT_OPT_GET_OWNER_VALIDATION,
+    GIT_OPT_GET_PACK_MAX_OBJECTS,
+    GIT_OPT_GET_SEARCH_PATH,
+    GIT_OPT_GET_SERVER_CONNECT_TIMEOUT,
+    GIT_OPT_GET_SERVER_TIMEOUT,
+    GIT_OPT_GET_TEMPLATE_PATH,
+    GIT_OPT_GET_USER_AGENT,
+    GIT_OPT_GET_USER_AGENT_PRODUCT,
+    GIT_OPT_GET_WINDOWS_SHAREMODE,
+    GIT_OPT_SET_ALLOCATOR,
+    GIT_OPT_SET_CACHE_MAX_SIZE,
+    GIT_OPT_SET_CACHE_OBJECT_LIMIT,
+    GIT_OPT_SET_EXTENSIONS,
+    GIT_OPT_SET_HOMEDIR,
+    GIT_OPT_SET_MWINDOW_FILE_LIMIT,
+    GIT_OPT_SET_MWINDOW_MAPPED_LIMIT,
+    GIT_OPT_SET_MWINDOW_SIZE,
+    GIT_OPT_SET_ODB_LOOSE_PRIORITY,
+    GIT_OPT_SET_ODB_PACKED_PRIORITY,
+    GIT_OPT_SET_OWNER_VALIDATION,
+    GIT_OPT_SET_PACK_MAX_OBJECTS,
+    GIT_OPT_SET_SEARCH_PATH,
+    GIT_OPT_SET_SERVER_CONNECT_TIMEOUT,
+    GIT_OPT_SET_SERVER_TIMEOUT,
+    GIT_OPT_SET_SSL_CERT_LOCATIONS,
+    GIT_OPT_SET_SSL_CIPHERS,
+    GIT_OPT_SET_TEMPLATE_PATH,
+    GIT_OPT_SET_USER_AGENT,
+    GIT_OPT_SET_USER_AGENT_PRODUCT,
+    GIT_OPT_SET_WINDOWS_SHAREMODE,
+    option,
+)
 from .packbuilder import PackBuilder
 from .remotes import Remote
 from .repository import Repository
@@ -60,7 +368,6 @@ from .settings import Settings
 from .submodules import Submodule
 from .utils import to_bytes, to_str
 
-
 # Features
 features = enums.Feature(C.git_libgit2_features())
 
@@ -73,12 +380,10 @@ _cache_enums()
 
 
 def init_repository(
-    path: typing.Union[str, bytes, os.PathLike, None],
+    path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None,
     bare: bool = False,
     flags: enums.RepositoryInitFlag = enums.RepositoryInitFlag.MKPATH,
-    mode: typing.Union[
-        int, enums.RepositoryInitMode
-    ] = enums.RepositoryInitMode.SHARED_UMASK,
+    mode: int | enums.RepositoryInitMode = enums.RepositoryInitMode.SHARED_UMASK,
     workdir_path: typing.Optional[str] = None,
     description: typing.Optional[str] = None,
     template_path: typing.Optional[str] = None,
@@ -155,8 +460,8 @@ def init_repository(
 
 
 def clone_repository(
-    url: str | bytes | os.PathLike,
-    path: str | bytes | os.PathLike,
+    url: str | bytes | os.PathLike[str] | os.PathLike[bytes],
+    path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
     bare: bool = False,
     repository: typing.Callable | None = None,
     remote: typing.Callable | None = None,
@@ -164,7 +469,7 @@ def clone_repository(
     callbacks: RemoteCallbacks | None = None,
     depth: int = 0,
     proxy: None | bool | str = None,
-):
+) -> Repository:
     """
     Clones a new Git repository from *url* in the given *path*.
 
@@ -456,37 +761,51 @@ __all__ = (
     'GIT_OBJECT_REF_DELTA',
     'GIT_OBJECT_TAG',
     'GIT_OBJECT_TREE',
+    'GIT_OPT_ADD_SSL_X509_CERT',
     'GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS',
     'GIT_OPT_ENABLE_CACHING',
     'GIT_OPT_ENABLE_FSYNC_GITDIR',
+    'GIT_OPT_ENABLE_HTTP_EXPECT_CONTINUE',
     'GIT_OPT_ENABLE_OFS_DELTA',
     'GIT_OPT_ENABLE_STRICT_HASH_VERIFICATION',
     'GIT_OPT_ENABLE_STRICT_OBJECT_CREATION',
     'GIT_OPT_ENABLE_STRICT_SYMBOLIC_REF_CREATION',
     'GIT_OPT_ENABLE_UNSAVED_INDEX_SAFETY',
     'GIT_OPT_GET_CACHED_MEMORY',
+    'GIT_OPT_GET_EXTENSIONS',
+    'GIT_OPT_GET_HOMEDIR',
     'GIT_OPT_GET_MWINDOW_FILE_LIMIT',
     'GIT_OPT_GET_MWINDOW_MAPPED_LIMIT',
     'GIT_OPT_GET_MWINDOW_SIZE',
     'GIT_OPT_GET_OWNER_VALIDATION',
     'GIT_OPT_GET_PACK_MAX_OBJECTS',
     'GIT_OPT_GET_SEARCH_PATH',
+    'GIT_OPT_GET_SERVER_CONNECT_TIMEOUT',
+    'GIT_OPT_GET_SERVER_TIMEOUT',
     'GIT_OPT_GET_TEMPLATE_PATH',
     'GIT_OPT_GET_USER_AGENT',
+    'GIT_OPT_GET_USER_AGENT_PRODUCT',
     'GIT_OPT_GET_WINDOWS_SHAREMODE',
     'GIT_OPT_SET_ALLOCATOR',
     'GIT_OPT_SET_CACHE_MAX_SIZE',
     'GIT_OPT_SET_CACHE_OBJECT_LIMIT',
+    'GIT_OPT_SET_EXTENSIONS',
+    'GIT_OPT_SET_HOMEDIR',
     'GIT_OPT_SET_MWINDOW_FILE_LIMIT',
     'GIT_OPT_SET_MWINDOW_MAPPED_LIMIT',
     'GIT_OPT_SET_MWINDOW_SIZE',
+    'GIT_OPT_SET_ODB_LOOSE_PRIORITY',
+    'GIT_OPT_SET_ODB_PACKED_PRIORITY',
     'GIT_OPT_SET_OWNER_VALIDATION',
     'GIT_OPT_SET_PACK_MAX_OBJECTS',
     'GIT_OPT_SET_SEARCH_PATH',
+    'GIT_OPT_SET_SERVER_CONNECT_TIMEOUT',
+    'GIT_OPT_SET_SERVER_TIMEOUT',
     'GIT_OPT_SET_SSL_CERT_LOCATIONS',
     'GIT_OPT_SET_SSL_CIPHERS',
     'GIT_OPT_SET_TEMPLATE_PATH',
     'GIT_OPT_SET_USER_AGENT',
+    'GIT_OPT_SET_USER_AGENT_PRODUCT',
     'GIT_OPT_SET_WINDOWS_SHAREMODE',
     'GIT_REFERENCES_ALL',
     'GIT_REFERENCES_BRANCHES',
