@@ -1,4 +1,4 @@
-# Copyright 2010-2024 The pygit2 contributors
+# Copyright 2010-2025 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -63,18 +63,26 @@ API.
 """
 
 # Standard Library
+from collections.abc import Callable, Generator
 from contextlib import contextmanager
 from functools import wraps
-from typing import Optional, Union
+from typing import TYPE_CHECKING, Optional, ParamSpec, TypeVar
 
 # pygit2
-from ._pygit2 import Oid, DiffFile
+from ._pygit2 import DiffFile, Oid
+from .credentials import Keypair, Username, UserPass
 from .enums import CheckoutNotify, CheckoutStrategy, CredentialType, StashApplyProgress
-from .errors import check_error, Passthrough
-from .ffi import ffi, C
-from .utils import maybe_string, to_bytes, ptr_to_bytes, StrArray
+from .errors import Passthrough, check_error
+from .ffi import C, ffi
+from .utils import StrArray, maybe_string, ptr_to_bytes, to_bytes
 
+_Credentials = Username | UserPass | Keypair
 
+if TYPE_CHECKING:
+    from pygit2._libgit2.ffi import GitProxyOptionsC
+
+    from ._pygit2 import CloneOptions, PushOptions
+    from .remotes import PushUpdate, TransferProgress
 #
 # The payload is the way to pass information from the pygit2 API, through
 # libgit2, to the Python callbacks. And back.
@@ -82,12 +90,16 @@ from .utils import maybe_string, to_bytes, ptr_to_bytes, StrArray
 
 
 class Payload:
-    def __init__(self, **kw):
+    repository: Callable | None
+    remote: Callable | None
+    clone_options: 'CloneOptions'
+
+    def __init__(self, **kw: object) -> None:
         for key, value in kw.items():
             setattr(self, key, value)
         self._stored_exception = None
 
-    def check_error(self, error_code):
+    def check_error(self, error_code: int) -> None:
         if error_code == C.GIT_EUSER:
             assert self._stored_exception is not None
             raise self._stored_exception
@@ -113,14 +125,20 @@ class RemoteCallbacks(Payload):
     RemoteCallbacks(certificate=certificate).
     """
 
-    def __init__(self, credentials=None, certificate=None):
+    push_options: 'PushOptions'
+
+    def __init__(
+        self,
+        credentials: _Credentials | None = None,
+        certificate_check: Callable[[None, bool, bytes], bool] | None = None,
+    ) -> None:
         super().__init__()
         if credentials is not None:
-            self.credentials = credentials
-        if certificate is not None:
-            self.certificate = certificate
+            self.credentials = credentials  # type: ignore[method-assign, assignment]
+        if certificate_check is not None:
+            self.certificate_check = certificate_check  # type: ignore[method-assign, assignment]
 
-    def sideband_progress(self, string):
+    def sideband_progress(self, string: str) -> None:
         """
         Progress output callback.  Override this function with your own
         progress reporting function
@@ -134,9 +152,9 @@ class RemoteCallbacks(Payload):
     def credentials(
         self,
         url: str,
-        username_from_url: Union[str, None],
+        username_from_url: str | None,
         allowed_types: CredentialType,
-    ):
+    ) -> _Credentials:
         """
         Credentials callback.  If the remote server requires authentication,
         this function will be called and its return value used for
@@ -159,7 +177,7 @@ class RemoteCallbacks(Payload):
         """
         raise Passthrough
 
-    def certificate_check(self, certificate, valid, host):
+    def certificate_check(self, certificate: None, valid: bool, host: bytes) -> bool:
         """
         Certificate callback. Override with your own function to determine
         whether to accept the server's certificate.
@@ -181,10 +199,21 @@ class RemoteCallbacks(Payload):
 
         raise Passthrough
 
-    def transfer_progress(self, stats):
+    def push_negotiation(self, updates: list['PushUpdate']) -> None:
         """
-        Transfer progress callback. Override with your own function to report
-        transfer progress.
+        During a push, called once between the negotiation step and the upload.
+        Provides information about what updates will be performed.
+
+        Override with your own function to check the pending updates
+        and possibly reject them (by raising an exception).
+        """
+
+    def transfer_progress(self, stats: 'TransferProgress') -> None:
+        """
+        During the download of new data, this will be regularly called with
+        the indexer's progress.
+
+        Override with your own function to report transfer progress.
 
         Parameters:
 
@@ -192,7 +221,20 @@ class RemoteCallbacks(Payload):
             The progress up to now.
         """
 
-    def update_tips(self, refname, old, new):
+    def push_transfer_progress(
+        self, objects_pushed: int, total_objects: int, bytes_pushed: int
+    ) -> None:
+        """
+        During the upload portion of a push, this will be regularly called
+        with progress information.
+
+        Be aware that this is called inline with pack building operations,
+        so performance may be affected.
+
+        Override with your own function to report push transfer progress.
+        """
+
+    def update_tips(self, refname: str, old: Oid, new: Oid) -> None:
         """
         Update tips callback. Override with your own function to report
         reference updates.
@@ -209,7 +251,7 @@ class RemoteCallbacks(Payload):
             The reference's new value.
         """
 
-    def push_update_reference(self, refname, message):
+    def push_update_reference(self, refname: str, message: str) -> None:
         """
         Push update reference callback. Override with your own function to
         report the remote's acceptance or rejection of reference updates.
@@ -229,7 +271,7 @@ class CheckoutCallbacks(Payload):
     in your class, which you can then pass to checkout operations.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__()
 
     def checkout_notify_flags(self) -> CheckoutNotify:
@@ -260,7 +302,7 @@ class CheckoutCallbacks(Payload):
         baseline: Optional[DiffFile],
         target: Optional[DiffFile],
         workdir: Optional[DiffFile],
-    ):
+    ) -> None:
         """
         Checkout will invoke an optional notification callback for
         certain cases - you pick which ones via `checkout_notify_flags`.
@@ -275,7 +317,9 @@ class CheckoutCallbacks(Payload):
         """
         pass
 
-    def checkout_progress(self, path: str, completed_steps: int, total_steps: int):
+    def checkout_progress(
+        self, path: str, completed_steps: int, total_steps: int
+    ) -> None:
         """
         Optional callback to notify the consumer of checkout progress.
         """
@@ -289,7 +333,7 @@ class StashApplyCallbacks(CheckoutCallbacks):
     in your class, which you can then pass to stash apply or pop operations.
     """
 
-    def stash_apply_progress(self, progress: StashApplyProgress):
+    def stash_apply_progress(self, progress: StashApplyProgress) -> None:
         """
         Stash application progress notification function.
 
@@ -344,7 +388,7 @@ def git_fetch_options(payload, opts=None):
     opts.callbacks.transfer_progress = C._transfer_progress_cb
     opts.callbacks.update_tips = C._update_tips_cb
     opts.callbacks.credentials = C._credentials_cb
-    opts.callbacks.certificate_check = C._certificate_cb
+    opts.callbacks.certificate_check = C._certificate_check_cb
     # Payload
     handle = ffi.new_handle(payload)
     opts.callbacks.payload = handle
@@ -355,6 +399,29 @@ def git_fetch_options(payload, opts=None):
     yield payload
 
 
+@contextmanager
+def git_proxy_options(
+    payload: object,
+    opts: Optional['GitProxyOptionsC'] = None,
+    proxy: None | bool | str = None,
+) -> Generator['GitProxyOptionsC', None, None]:
+    if opts is None:
+        opts = ffi.new('git_proxy_options *')
+        C.git_proxy_options_init(opts, C.GIT_PROXY_OPTIONS_VERSION)
+    if proxy is None:
+        opts.type = C.GIT_PROXY_NONE
+    elif proxy is True:
+        opts.type = C.GIT_PROXY_AUTO
+    elif type(proxy) is str:
+        opts.type = C.GIT_PROXY_SPECIFIED
+        # Keep url in memory, otherwise memory is freed and bad things happen
+        payload.__proxy_url = ffi.new('char[]', to_bytes(proxy))  # type: ignore[attr-defined]
+        opts.url = payload.__proxy_url  # type: ignore[attr-defined]
+    else:
+        raise TypeError('Proxy must be None, True, or a string')
+    yield opts
+
+
 @contextmanager
 def git_push_options(payload, opts=None):
     if payload is None:
@@ -368,8 +435,16 @@ def git_push_options(payload, opts=None):
     opts.callbacks.transfer_progress = C._transfer_progress_cb
     opts.callbacks.update_tips = C._update_tips_cb
     opts.callbacks.credentials = C._credentials_cb
-    opts.callbacks.certificate_check = C._certificate_cb
+    opts.callbacks.certificate_check = C._certificate_check_cb
     opts.callbacks.push_update_reference = C._push_update_reference_cb
+    opts.callbacks.push_negotiation = C._push_negotiation_cb
+    # Per libgit2 sources, push_transfer_progress may incur a performance hit.
+    # So, set it only if the user has overridden the no-op stub.
+    if (
+        type(payload).push_transfer_progress
+        is not RemoteCallbacks.push_transfer_progress
+    ):
+        opts.callbacks.push_transfer_progress = C._push_transfer_progress_cb
     # Payload
     handle = ffi.new_handle(payload)
     opts.callbacks.payload = handle
@@ -391,6 +466,7 @@ def git_remote_callbacks(payload):
     # Plug callbacks
     cdata.credentials = C._credentials_cb
     cdata.update_tips = C._update_tips_cb
+    cdata.certificate_check = C._certificate_check_cb
     # Payload
     handle = ffi.new_handle(payload)
     cdata.payload = handle
@@ -404,18 +480,21 @@ def git_remote_callbacks(payload):
 #
 # C callbacks
 #
-# These functions are called by libgit2. They cannot raise execptions, since
+# These functions are called by libgit2. They cannot raise exceptions, since
 # they return to libgit2, they can only send back error codes.
 #
-# They cannot be overriden, but sometimes the only thing these functions do is
+# They cannot be overridden, but sometimes the only thing these functions do is
 # to proxy the call to a user defined function. If user defined functions
 # raises an exception, the callback must store it somewhere and return
 # GIT_EUSER to libgit2, then the outer Python code will be able to reraise the
 # exception.
 #
 
+P = ParamSpec('P')
+T = TypeVar('T')
 
-def libgit2_callback(f):
+
+def libgit2_callback(f: Callable[P, T]) -> Callable[P, T]:
     @wraps(f)
     def wrapper(*args):
         data = ffi.from_handle(args[-1])
@@ -433,10 +512,10 @@ def libgit2_callback(f):
             data._stored_exception = e
             return C.GIT_EUSER
 
-    return ffi.def_extern()(wrapper)
+    return ffi.def_extern()(wrapper)  # type: ignore[attr-defined]
 
 
-def libgit2_callback_void(f):
+def libgit2_callback_void(f: Callable[P, T]) -> Callable[P, T]:
     @wraps(f)
     def wrapper(*args):
         data = ffi.from_handle(args[-1])
@@ -453,11 +532,11 @@ def libgit2_callback_void(f):
             data._stored_exception = e
             pass  # Function returns void, so we can't do much here.
 
-    return ffi.def_extern()(wrapper)
+    return ffi.def_extern()(wrapper)  # type: ignore[attr-defined]
 
 
 @libgit2_callback
-def _certificate_cb(cert_i, valid, host, data):
+def _certificate_check_cb(cert_i, valid, host, data):
     # We want to simulate what should happen if libgit2 supported pass-through
     # for this callback. For SSH, 'valid' is always False, because it doesn't
     # look at known_hosts, but we do want to let it through in order to do what
@@ -494,6 +573,19 @@ def _credentials_cb(cred_out, url, username, allowed, data):
     return 0
 
 
+@libgit2_callback
+def _push_negotiation_cb(updates, num_updates, data):
+    from .remotes import PushUpdate
+
+    push_negotiation = getattr(data, 'push_negotiation', None)
+    if not push_negotiation:
+        return 0
+
+    py_updates = [PushUpdate(updates[i]) for i in range(num_updates)]
+    push_negotiation(py_updates)
+    return 0
+
+
 @libgit2_callback
 def _push_update_reference_cb(ref, msg, data):
     push_update_reference = getattr(data, 'push_update_reference', None)
@@ -553,6 +645,16 @@ def _transfer_progress_cb(stats_ptr, data):
     return 0
 
 
+@libgit2_callback
+def _push_transfer_progress_cb(current, total, bytes_pushed, payload):
+    push_transfer_progress = getattr(payload, 'push_transfer_progress', None)
+    if not push_transfer_progress:
+        return 0
+
+    push_transfer_progress(current, total, bytes_pushed)
+    return 0
+
+
 @libgit2_callback
 def _update_tips_cb(refname, a, b, data):
     update_tips = getattr(data, 'update_tips', None)
@@ -643,7 +745,7 @@ def _checkout_notify_cb(
     pyworkdir = DiffFile.from_c(ptr_to_bytes(workdir))
 
     try:
-        data.checkout_notify(why, pypath, pybaseline, pytarget, pyworkdir)
+        data.checkout_notify(why, pypath, pybaseline, pytarget, pyworkdir)  # type: ignore[arg-type]
     except Passthrough:
         # Unlike most other operations with optional callbacks, checkout
         # doesn't support the GIT_PASSTHROUGH return code, so we must bypass
@@ -658,7 +760,7 @@ def _checkout_notify_cb(
 
 @libgit2_callback_void
 def _checkout_progress_cb(path, completed_steps, total_steps, data: CheckoutCallbacks):
-    data.checkout_progress(maybe_string(path), completed_steps, total_steps)
+    data.checkout_progress(maybe_string(path), completed_steps, total_steps)  # type: ignore[arg-type]
 
 
 def _git_checkout_options(
