@@ -23,27 +23,43 @@
 # the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 # Boston, MA 02110-1301, USA.
 
+import tarfile
+import warnings
+from collections.abc import Callable, Iterator
 from io import BytesIO
-from os import PathLike
+from pathlib import Path
 from string import hexdigits
 from time import time
-import tarfile
-import typing
+from typing import TYPE_CHECKING, Optional, overload
 
 # Import from pygit2
-from ._pygit2 import Repository as _Repository, init_file_backend
-from ._pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
-from ._pygit2 import Reference, Tree, Commit, Blob, Signature
-from ._pygit2 import InvalidSpecError
-
+from ._pygit2 import (
+    GIT_OID_HEXSZ,
+    GIT_OID_MINPREFIXLEN,
+    Blob,
+    Commit,
+    Diff,
+    InvalidSpecError,
+    Object,
+    Oid,
+    Patch,
+    Reference,
+    Signature,
+    Tree,
+    init_file_backend,
+)
+from ._pygit2 import Repository as _Repository
 from .blame import Blame
 from .branches import Branches
-from .callbacks import git_checkout_options, git_stash_apply_options
+from .callbacks import (
+    StashApplyCallbacks,
+    git_checkout_options,
+    git_stash_apply_options,
+)
 from .config import Config
 from .enums import (
     AttrCheck,
     BlameFlag,
-    BranchType,
     CheckoutStrategy,
     DescribeStrategy,
     DiffOption,
@@ -56,21 +72,50 @@ from .enums import (
     RepositoryState,
 )
 from .errors import check_error
-from .ffi import ffi, C
-from .index import Index, IndexEntry
+from .ffi import C, ffi
+from .index import Index, IndexEntry, MergeFileResult
 from .packbuilder import PackBuilder
 from .references import References
 from .remotes import RemoteCollection
 from .submodules import SubmoduleCollection
-from .utils import to_bytes, StrArray
+from .utils import StrArray, to_bytes
+
+if TYPE_CHECKING:
+    from pygit2._libgit2.ffi import (
+        ArrayC,
+        GitMergeOptionsC,
+        GitRepositoryC,
+        _Pointer,
+        char,
+    )
+    from pygit2._pygit2 import Odb, Refdb, RefdbBackend
 
 
 class BaseRepository(_Repository):
-    def __init__(self, *args, **kwargs):
+    _pointer: '_Pointer[GitRepositoryC]'
+    _repo: 'GitRepositoryC'
+    backend: 'RefdbBackend'
+    default_signature: Signature
+    head: Reference
+    head_is_detached: bool
+    head_is_unborn: bool
+    is_bare: bool
+    is_empty: bool
+    is_shallow: bool
+    odb: 'Odb'
+    path: str
+    refdb: 'Refdb'
+    workdir: str
+    references: References
+    remotes: RemoteCollection
+    branches: Branches
+    submodules: SubmoduleCollection
+
+    def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
         self._common_init()
 
-    def _common_init(self):
+    def _common_init(self) -> None:
         self.branches = Branches(self)
         self.references = References(self)
         self.remotes = RemoteCollection(self)
@@ -83,22 +128,27 @@ class BaseRepository(_Repository):
         self._repo = repo_cptr[0]
 
     # Backwards compatible ODB access
-    def read(self, *args, **kwargs):
+    def read(self, oid: Oid | str) -> tuple[int, bytes]:
         """read(oid) -> type, data, size
 
         Read raw object data from the repository.
         """
-        return self.odb.read(*args, **kwargs)
+        return self.odb.read(oid)
 
-    def write(self, *args, **kwargs):
+    def write(self, type: int, data: bytes | str) -> Oid:
         """write(type, data) -> Oid
 
         Write raw object data into the repository. First arg is the object
         type, the second one a buffer with data. Return the Oid of the created
         object."""
-        return self.odb.write(*args, **kwargs)
+        return self.odb.write(type, data)
 
-    def pack(self, path=None, pack_delegate=None, n_threads=None):
+    def pack(
+        self,
+        path: str | Path | None = None,
+        pack_delegate: Callable[[PackBuilder], None] | None = None,
+        n_threads: int | None = None,
+    ) -> int:
         """Pack the objects in the odb chosen by the pack_delegate function
         and write `.pack` and `.idx` files for them.
 
@@ -134,8 +184,8 @@ class BaseRepository(_Repository):
         self,
         path: str,
         object_type: ObjectType = ObjectType.BLOB,
-        as_path: typing.Optional[str] = None,
-    ):
+        as_path: str | None = None,
+    ) -> Oid:
         """Calculate the hash of a file using repository filtering rules.
 
         If you simply want to calculate the hash of a file on disk with no filters,
@@ -167,6 +217,7 @@ class BaseRepository(_Repository):
         """
         c_path = to_bytes(path)
 
+        c_as_path: ffi.NULL_TYPE | bytes
         if as_path is None:
             c_as_path = ffi.NULL
         else:
@@ -182,33 +233,33 @@ class BaseRepository(_Repository):
         oid = Oid(raw=bytes(ffi.buffer(c_oid.id)[:]))
         return oid
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Oid]:
         return iter(self.odb)
 
     #
     # Mapping interface
     #
-    def get(self, key, default=None):
+    def get(self, key: Oid | str, default: Optional[Commit] = None) -> None | Object:
         value = self.git_object_lookup_prefix(key)
         return value if (value is not None) else default
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: str | Oid) -> Object:
         value = self.git_object_lookup_prefix(key)
         if value is None:
             raise KeyError(key)
         return value
 
-    def __contains__(self, key):
+    def __contains__(self, key: str | Oid) -> bool:
         return self.git_object_lookup_prefix(key) is not None
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f'pygit2.Repository({repr(self.path)})'
 
     #
     # Configuration
     #
     @property
-    def config(self):
+    def config(self) -> Config:
         """The configuration file for this repository.
 
         If a the configuration hasn't been set yet, the default config for
@@ -237,7 +288,13 @@ class BaseRepository(_Repository):
     #
     # References
     #
-    def create_reference(self, name, target, force=False, message=None):
+    def create_reference(
+        self,
+        name: str,
+        target: Oid | str,
+        force: bool = False,
+        message: str | None = None,
+    ) -> 'Reference':
         """Create a new reference "name" which points to an object or to
         another reference.
 
@@ -259,25 +316,26 @@ class BaseRepository(_Repository):
             repo.create_reference('refs/tags/foo', 'refs/heads/master')
             repo.create_reference('refs/tags/foo', 'bbb78a9cec580')
         """
-        direct = type(target) is Oid or (
+        direct = isinstance(target, Oid) or (
             all(c in hexdigits for c in target)
             and GIT_OID_MINPREFIXLEN <= len(target) <= GIT_OID_HEXSZ
         )
 
-        if direct:
+        # duplicate isinstance call for mypy
+        if direct or isinstance(target, Oid):
             return self.create_reference_direct(name, target, force, message=message)
 
         return self.create_reference_symbolic(name, target, force, message=message)
 
-    def listall_references(self) -> typing.List[str]:
+    def listall_references(self) -> list[str]:
         """Return a list with all the references in the repository."""
         return list(x.name for x in self.references.iterator())
 
-    def listall_reference_objects(self) -> typing.List[Reference]:
+    def listall_reference_objects(self) -> list[Reference]:
         """Return a list with all the reference objects in the repository."""
         return list(x for x in self.references.iterator())
 
-    def resolve_refish(self, refish):
+    def resolve_refish(self, refish: str) -> tuple[Commit, Reference]:
         """Convert a reference-like short name "ref-ish" to a valid
         (commit, reference) pair.
 
@@ -297,9 +355,9 @@ class BaseRepository(_Repository):
             reference = None
             commit = self.revparse_single(refish)
         else:
-            commit = reference.peel(Commit)
+            commit = reference.peel(Commit)  # type: ignore
 
-        return (commit, reference)
+        return (commit, reference)  # type: ignore
 
     #
     # Checkout
@@ -338,7 +396,11 @@ class BaseRepository(_Repository):
             err = C.git_checkout_tree(self._repo, cptr[0], payload.checkout_options)
             payload.check_error(err)
 
-    def checkout(self, refname=None, **kwargs):
+    def checkout(
+        self,
+        refname: str | None | Reference = None,
+        **kwargs,
+    ) -> None:
         """
         Checkout the given reference using the given strategy, and update the
         HEAD.
@@ -406,7 +468,7 @@ class BaseRepository(_Repository):
     #
     # Setting HEAD
     #
-    def set_head(self, target):
+    def set_head(self, target: Oid | str) -> None:
         """
         Set HEAD to point to the given target.
 
@@ -453,15 +515,35 @@ class BaseRepository(_Repository):
 
         return obj
 
+    @overload
     def diff(
         self,
-        a=None,
-        b=None,
-        cached=False,
+        a: None | str | bytes | Commit | Oid | Reference = None,
+        b: None | str | bytes | Commit | Oid | Reference = None,
+        cached: bool = False,
         flags: DiffOption = DiffOption.NORMAL,
         context_lines: int = 3,
         interhunk_lines: int = 0,
-    ):
+    ) -> Diff: ...
+    @overload
+    def diff(
+        self,
+        a: Blob | None = None,
+        b: Blob | None = None,
+        cached: bool = False,
+        flags: DiffOption = DiffOption.NORMAL,
+        context_lines: int = 3,
+        interhunk_lines: int = 0,
+    ) -> Patch: ...
+    def diff(
+        self,
+        a: None | Blob | str | bytes | Commit | Oid | Reference = None,
+        b: None | Blob | str | bytes | Commit | Oid | Reference = None,
+        cached: bool = False,
+        flags: DiffOption = DiffOption.NORMAL,
+        context_lines: int = 3,
+        interhunk_lines: int = 0,
+    ) -> Diff | Patch:
         """
         Show changes between the working tree and the index or a tree,
         changes between the index and a tree, changes between two trees, or
@@ -522,27 +604,30 @@ class BaseRepository(_Repository):
         a = self.__whatever_to_tree_or_blob(a)
         b = self.__whatever_to_tree_or_blob(b)
 
-        opt_keys = ['flags', 'context_lines', 'interhunk_lines']
-        opt_values = [int(flags), context_lines, interhunk_lines]
+        options = {
+            'flags': int(flags),
+            'context_lines': context_lines,
+            'interhunk_lines': interhunk_lines,
+        }
 
         # Case 1: Diff tree to tree
         if isinstance(a, Tree) and isinstance(b, Tree):
-            return a.diff_to_tree(b, **dict(zip(opt_keys, opt_values)))
+            return a.diff_to_tree(b, **options)  # type: ignore[arg-type]
 
         # Case 2: Index to workdir
         elif a is None and b is None:
-            return self.index.diff_to_workdir(*opt_values)
+            return self.index.diff_to_workdir(**options)
 
         # Case 3: Diff tree to index or workdir
         elif isinstance(a, Tree) and b is None:
             if cached:
-                return a.diff_to_index(self.index, *opt_values)
+                return a.diff_to_index(self.index, **options)  # type: ignore[arg-type]
             else:
-                return a.diff_to_workdir(*opt_values)
+                return a.diff_to_workdir(**options)  # type: ignore[arg-type]
 
         # Case 4: Diff blob to blob
         if isinstance(a, Blob) and isinstance(b, Blob):
-            return a.diff(b)
+            return a.diff(b, **options)  # type: ignore[arg-type]
 
         raise ValueError('Only blobs and treeish can be diffed')
 
@@ -557,9 +642,9 @@ class BaseRepository(_Repository):
             return RepositoryState(cstate)
         except ValueError:
             # Some value not in the IntEnum - newer libgit2 version?
-            return cstate
+            return cstate  # type: ignore[return-value]
 
-    def state_cleanup(self):
+    def state_cleanup(self) -> None:
         """Remove all the metadata associated with an ongoing command like
         merge, revert, cherry-pick, etc. For example: MERGE_HEAD, MERGE_MSG,
         etc.
@@ -571,14 +656,14 @@ class BaseRepository(_Repository):
     #
     def blame(
         self,
-        path,
+        path: str,
         flags: BlameFlag = BlameFlag.NORMAL,
-        min_match_characters=None,
-        newest_commit=None,
-        oldest_commit=None,
-        min_line=None,
-        max_line=None,
-    ):
+        min_match_characters: int | None = None,
+        newest_commit: Oid | str | None = None,
+        oldest_commit: Oid | str | None = None,
+        min_line: int | None = None,
+        max_line: int | None = None,
+    ) -> Blame:
         """
         Return a Blame object for a single file.
 
@@ -612,6 +697,7 @@ class BaseRepository(_Repository):
         """
 
         options = ffi.new('git_blame_options *')
+
         C.git_blame_options_init(options, C.GIT_BLAME_OPTIONS_VERSION)
         if flags:
             options.flags = int(flags)
@@ -652,7 +738,9 @@ class BaseRepository(_Repository):
     # Merging
     #
     @staticmethod
-    def _merge_options(favor: MergeFavor, flags: MergeFlag, file_flags: MergeFileFlag):
+    def _merge_options(
+        favor: int | MergeFavor, flags: int | MergeFlag, file_flags: int | MergeFileFlag
+    ) -> 'GitMergeOptionsC':
         """Return a 'git_merge_opts *'"""
 
         # Check arguments type
@@ -677,12 +765,16 @@ class BaseRepository(_Repository):
 
     def merge_file_from_index(
         self,
-        ancestor: typing.Union[None, IndexEntry],
-        ours: typing.Union[None, IndexEntry],
-        theirs: typing.Union[None, IndexEntry],
-    ) -> str:
-        """Merge files from index. Return a string with the merge result
-        containing possible conflicts.
+        ancestor: 'IndexEntry | None',
+        ours: 'IndexEntry | None',
+        theirs: 'IndexEntry | None',
+        use_deprecated: bool = True,
+    ) -> 'str | MergeFileResult | None':
+        """Merge files from index.
+
+        Returns: A string with the content of the file containing
+        possible conflicts if use_deprecated==True.
+        If use_deprecated==False then it returns an instance of MergeFileResult.
 
         ancestor
             The index entry which will be used as a common
@@ -691,6 +783,10 @@ class BaseRepository(_Repository):
             The index entry to take as "ours" or base.
         theirs
             The index entry which will be merged into "ours"
+        use_deprecated
+            This controls what will be returned. If use_deprecated==True (default),
+            a string with the contents of the file will be returned.
+            An instance of MergeFileResult will be returned otherwise.
         """
         cmergeresult = ffi.new('git_merge_file_result *')
 
@@ -707,19 +803,28 @@ class BaseRepository(_Repository):
         )
         check_error(err)
 
-        ret = ffi.string(cmergeresult.ptr, cmergeresult.len).decode('utf-8')
+        mergeFileResult = MergeFileResult._from_c(cmergeresult)
         C.git_merge_file_result_free(cmergeresult)
 
-        return ret
+        if use_deprecated:
+            warnings.warn(
+                'Getting an str from Repository.merge_file_from_index is deprecated. '
+                'The method will later return an instance of MergeFileResult by default, instead. '
+                'Check parameter use_deprecated.',
+                DeprecationWarning,
+            )
+            return mergeFileResult.contents if mergeFileResult else ''
+
+        return mergeFileResult
 
     def merge_commits(
         self,
-        ours: typing.Union[str, Oid, Commit],
-        theirs: typing.Union[str, Oid, Commit],
-        favor=MergeFavor.NORMAL,
-        flags=MergeFlag.FIND_RENAMES,
-        file_flags=MergeFileFlag.DEFAULT,
-    ) -> Index:
+        ours: str | Oid | Commit,
+        theirs: str | Oid | Commit,
+        favor: MergeFavor = MergeFavor.NORMAL,
+        flags: MergeFlag = MergeFlag.FIND_RENAMES,
+        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
+    ) -> 'Index':
         """
         Merge two arbitrary commits.
 
@@ -751,9 +856,15 @@ class BaseRepository(_Repository):
         cindex = ffi.new('git_index **')
 
         if isinstance(ours, (str, Oid)):
-            ours = self[ours]
+            ours_object = self[ours]
+            if not isinstance(ours_object, Commit):
+                raise TypeError(f'expected Commit, got {type(ours_object)}')
+            ours = ours_object
         if isinstance(theirs, (str, Oid)):
-            theirs = self[theirs]
+            theirs_object = self[theirs]
+            if not isinstance(theirs_object, Commit):
+                raise TypeError(f'expected Commit, got {type(theirs_object)}')
+            theirs = theirs_object
 
         ours = ours.peel(Commit)
         theirs = theirs.peel(Commit)
@@ -770,13 +881,13 @@ class BaseRepository(_Repository):
 
     def merge_trees(
         self,
-        ancestor: typing.Union[str, Oid, Tree],
-        ours: typing.Union[str, Oid, Tree],
-        theirs: typing.Union[str, Oid, Tree],
-        favor=MergeFavor.NORMAL,
-        flags=MergeFlag.FIND_RENAMES,
-        file_flags=MergeFileFlag.DEFAULT,
-    ):
+        ancestor: str | Oid | Tree,
+        ours: str | Oid | Tree,
+        theirs: str | Oid | Tree,
+        favor: MergeFavor = MergeFavor.NORMAL,
+        flags: MergeFlag = MergeFlag.FIND_RENAMES,
+        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
+    ) -> 'Index':
         """
         Merge two trees.
 
@@ -808,16 +919,9 @@ class BaseRepository(_Repository):
         theirs_ptr = ffi.new('git_tree **')
         cindex = ffi.new('git_index **')
 
-        if isinstance(ancestor, (str, Oid)):
-            ancestor = self[ancestor]
-        if isinstance(ours, (str, Oid)):
-            ours = self[ours]
-        if isinstance(theirs, (str, Oid)):
-            theirs = self[theirs]
-
-        ancestor = ancestor.peel(Tree)
-        ours = ours.peel(Tree)
-        theirs = theirs.peel(Tree)
+        ancestor = self.__ensure_tree(ancestor)
+        ours = self.__ensure_tree(ours)
+        theirs = self.__ensure_tree(theirs)
 
         opts = self._merge_options(favor, flags, file_flags)
 
@@ -834,23 +938,28 @@ class BaseRepository(_Repository):
 
     def merge(
         self,
-        id: typing.Union[Oid, str],
-        favor=MergeFavor.NORMAL,
-        flags=MergeFlag.FIND_RENAMES,
-        file_flags=MergeFileFlag.DEFAULT,
-    ):
+        source: Reference | Commit | Oid | str,
+        favor: MergeFavor = MergeFavor.NORMAL,
+        flags: MergeFlag = MergeFlag.FIND_RENAMES,
+        file_flags: MergeFileFlag = MergeFileFlag.DEFAULT,
+    ) -> None:
         """
-        Merges the given id into HEAD.
+        Merges the given Reference or Commit into HEAD.
 
-        Merges the given commit(s) into HEAD, writing the results into the working directory.
+        Merges the given commit into HEAD, writing the results into the working directory.
         Any changes are staged for commit and any conflicts are written to the index.
         Callers should inspect the repository's index after this completes,
         resolve any conflicts and prepare a commit.
 
         Parameters:
 
-        id
-            The id to merge into HEAD
+        source
+            The Reference, Commit, or commit Oid to merge into HEAD.
+            It is preferable to pass in a Reference, because this enriches the
+            merge with additional information (for example, Repository.message will
+            specify the name of the branch being merged).
+            Previous versions of pygit2 allowed passing in a partial commit
+            hash as a string; this is deprecated.
 
         favor
             An enums.MergeFavor constant specifying how to deal with file-level conflicts.
@@ -862,12 +971,35 @@ class BaseRepository(_Repository):
         file_flags
             A combination of enums.MergeFileFlag constants.
         """
-        if not isinstance(id, (str, Oid)):
-            raise TypeError(f'expected oid (string or <Oid>) got {type(id)}')
 
-        id = self[id].id
-        c_id = ffi.new('git_oid *')
-        ffi.buffer(c_id)[:] = id.raw[:]
+        if isinstance(source, Reference):
+            # Annotated commit from ref
+            cptr = ffi.new('struct git_reference **')
+            ffi.buffer(cptr)[:] = source._pointer[:]  # type: ignore[attr-defined]
+            commit_ptr = ffi.new('git_annotated_commit **')
+            err = C.git_annotated_commit_from_ref(commit_ptr, self._repo, cptr[0])
+            check_error(err)
+        else:
+            # Annotated commit from commit id
+            if isinstance(source, str):
+                # For backwards compatibility, parse a string as a partial commit hash
+                warnings.warn(
+                    'Passing str to Repository.merge is deprecated. '
+                    'Pass Commit, Oid, or a Reference (such as a Branch) instead.',
+                    DeprecationWarning,
+                )
+                oid = self[source].peel(Commit).id
+            elif isinstance(source, Commit):
+                oid = source.id
+            elif isinstance(source, Oid):
+                oid = source
+            else:
+                raise TypeError('expected Reference, Commit, or Oid')
+            c_id = ffi.new('git_oid *')
+            ffi.buffer(c_id)[:] = oid.raw[:]
+            commit_ptr = ffi.new('git_annotated_commit **')
+            err = C.git_annotated_commit_lookup(commit_ptr, self._repo, c_id)
+            check_error(err)
 
         merge_opts = self._merge_options(favor, flags, file_flags)
 
@@ -877,10 +1009,6 @@ class BaseRepository(_Repository):
             CheckoutStrategy.SAFE | CheckoutStrategy.RECREATE_MISSING
         )
 
-        commit_ptr = ffi.new('git_annotated_commit **')
-        err = C.git_annotated_commit_lookup(commit_ptr, self._repo, c_id)
-        check_error(err)
-
         err = C.git_merge(self._repo, commit_ptr, 1, merge_opts, checkout_opts)
         C.git_annotated_commit_free(commit_ptr[0])
         check_error(err)
@@ -923,7 +1051,7 @@ class BaseRepository(_Repository):
         """
         return self.raw_message.decode('utf-8')
 
-    def remove_message(self):
+    def remove_message(self) -> None:
         """
         Remove git's prepared message.
         """
@@ -935,16 +1063,16 @@ class BaseRepository(_Repository):
     #
     def describe(
         self,
-        committish=None,
-        max_candidates_tags=None,
+        committish: str | Reference | Commit | None = None,
+        max_candidates_tags: int | None = None,
         describe_strategy: DescribeStrategy = DescribeStrategy.DEFAULT,
-        pattern=None,
-        only_follow_first_parent=None,
-        show_commit_oid_as_fallback=None,
-        abbreviated_size=None,
-        always_use_long_format=None,
-        dirty_suffix=None,
-    ):
+        pattern: str | None = None,
+        only_follow_first_parent: bool | None = None,
+        show_commit_oid_as_fallback: bool | None = None,
+        abbreviated_size: int | None = None,
+        always_use_long_format: bool | None = None,
+        dirty_suffix: str | None = None,
+    ) -> str:
         """
         Describe a commit-ish or the current working tree.
 
@@ -987,7 +1115,7 @@ class BaseRepository(_Repository):
 
         always_use_long_format : bool
             Always output the long format (the nearest tag, the number of
-            commits, and the abbrevated commit name) even when the committish
+            commits, and the abbreviated commit name) even when the committish
             matches a tag.
 
         dirty_suffix : str
@@ -1018,10 +1146,13 @@ class BaseRepository(_Repository):
 
         result = ffi.new('git_describe_result **')
         if committish:
+            committish_rev: Object | Reference | Commit
             if isinstance(committish, str):
-                committish = self.revparse_single(committish)
+                committish_rev = self.revparse_single(committish)
+            else:
+                committish_rev = committish
 
-            commit = committish.peel(Commit)
+            commit = committish_rev.peel(Commit)
 
             cptr = ffi.new('git_object **')
             ffi.buffer(cptr)[:] = commit._pointer[:]
@@ -1064,13 +1195,13 @@ class BaseRepository(_Repository):
     def stash(
         self,
         stasher: Signature,
-        message: typing.Optional[str] = None,
+        message: str | None = None,
         keep_index: bool = False,
         include_untracked: bool = False,
         include_ignored: bool = False,
         keep_all: bool = False,
-        paths: typing.Optional[typing.List[str]] = None,
-    ):
+        paths: list[str] | None = None,
+    ) -> Oid:
         """
         Save changes to the working directory to the stash.
 
@@ -1125,7 +1256,7 @@ class BaseRepository(_Repository):
 
         if paths:
             arr = StrArray(paths)
-            opts.paths = arr.ptr[0]
+            opts.paths = arr.ptr[0]  # type: ignore[index]
 
         coid = ffi.new('git_oid *')
         err = C.git_stash_save_with_opts(coid, self._repo, opts)
@@ -1134,7 +1265,13 @@ class BaseRepository(_Repository):
 
         return Oid(raw=bytes(ffi.buffer(coid)[:]))
 
-    def stash_apply(self, index=0, **kwargs):
+    def stash_apply(
+        self,
+        index: int = 0,
+        reinstate_index: bool = False,
+        strategy: CheckoutStrategy | None = None,
+        callbacks: StashApplyCallbacks | None = None,
+    ) -> None:
         """
         Apply a stashed state in the stash list to the working directory.
 
@@ -1168,11 +1305,15 @@ class BaseRepository(_Repository):
             >>> repo.stash(repo.default_signature(), 'WIP: stashing')
             >>> repo.stash_apply(strategy=CheckoutStrategy.ALLOW_CONFLICTS)
         """
-        with git_stash_apply_options(**kwargs) as payload:
+        with git_stash_apply_options(
+            reinstate_index=reinstate_index,
+            strategy=strategy,
+            callbacks=callbacks,
+        ) as payload:
             err = C.git_stash_apply(self._repo, index, payload.stash_apply_options)
             payload.check_error(err)
 
-    def stash_drop(self, index=0):
+    def stash_drop(self, index: int = 0) -> None:
         """
         Remove a stashed state from the stash list.
 
@@ -1184,19 +1325,35 @@ class BaseRepository(_Repository):
         """
         check_error(C.git_stash_drop(self._repo, index))
 
-    def stash_pop(self, index=0, **kwargs):
+    def stash_pop(
+        self,
+        index: int = 0,
+        reinstate_index: bool = False,
+        strategy: CheckoutStrategy | None = None,
+        callbacks: StashApplyCallbacks | None = None,
+    ) -> None:
         """Apply a stashed state and remove it from the stash list.
 
         For arguments, see Repository.stash_apply().
         """
-        with git_stash_apply_options(**kwargs) as payload:
+        with git_stash_apply_options(
+            reinstate_index=reinstate_index,
+            strategy=strategy,
+            callbacks=callbacks,
+        ) as payload:
             err = C.git_stash_pop(self._repo, index, payload.stash_apply_options)
             payload.check_error(err)
 
     #
     # Utility for writing a tree into an archive
     #
-    def write_archive(self, treeish, archive, timestamp=None, prefix=''):
+    def write_archive(
+        self,
+        treeish: str | Tree | Object | Oid,
+        archive: tarfile.TarFile,
+        timestamp: int | None = None,
+        prefix: str = '',
+    ) -> None:
         """
         Write treeish into an archive.
 
@@ -1268,7 +1425,7 @@ class BaseRepository(_Repository):
     #
     # Ahead-behind, which mostly lives on its own namespace
     #
-    def ahead_behind(self, local, upstream):
+    def ahead_behind(self, local: Oid | str, upstream: Oid | str) -> tuple[int, int]:
         """
         Calculate how many different commits are in the non-common parts of the
         history between the two given ids.
@@ -1308,11 +1465,11 @@ class BaseRepository(_Repository):
     #
     def get_attr(
         self,
-        path: typing.Union[str, bytes, PathLike],
-        name: typing.Union[str, bytes],
+        path: str | bytes | Path,
+        name: str | bytes,
         flags: AttrCheck = AttrCheck.FILE_THEN_INDEX,
-        commit: typing.Union[Oid, str, None] = None,
-    ) -> typing.Union[bool, None, str]:
+        commit: Oid | str | None = None,
+    ) -> bool | None | str:
         """
         Retrieve an attribute for a file by path.
 
@@ -1385,7 +1542,7 @@ class BaseRepository(_Repository):
 
         return (ffi.string(cname).decode('utf-8'), ffi.string(cemail).decode('utf-8'))
 
-    def set_ident(self, name, email):
+    def set_ident(self, name: str, email: str) -> None:
         """Set the identity to be used for reference operations.
 
         Updates to some references also append data to their
@@ -1396,7 +1553,7 @@ class BaseRepository(_Repository):
         err = C.git_repository_set_ident(self._repo, to_bytes(name), to_bytes(email))
         check_error(err)
 
-    def revert(self, commit: Commit):
+    def revert(self, commit: Commit) -> None:
         """
         Revert the given commit, producing changes in the index and working
         directory.
@@ -1409,7 +1566,9 @@ class BaseRepository(_Repository):
         err = C.git_revert(self._repo, commit_ptr[0], ffi.NULL)
         check_error(err)
 
-    def revert_commit(self, revert_commit, our_commit, mainline=0):
+    def revert_commit(
+        self, revert_commit: Commit, our_commit: Commit, mainline: int = 0
+    ) -> Index:
         """
         Revert the given Commit against the given "our" Commit, producing an
         Index that reflects the result of the revert.
@@ -1450,14 +1609,14 @@ class BaseRepository(_Repository):
     #
     def amend_commit(
         self,
-        commit,
-        refname,
-        author=None,
-        committer=None,
-        message=None,
-        tree=None,
-        encoding='UTF-8',
-    ):
+        commit: Commit | Oid | str,
+        refname: Reference | str | None,
+        author: Signature | None = None,
+        committer: Signature | None = None,
+        message: str | None = None,
+        tree: Tree | Oid | str | None = None,
+        encoding: str = 'UTF-8',
+    ) -> Oid:
         """
         Amend an existing commit by replacing only explicitly passed values,
         return the rewritten commit's oid.
@@ -1506,24 +1665,24 @@ class BaseRepository(_Repository):
         # Note: the pointers are all initialized to NULL by default.
         coid = ffi.new('git_oid *')
         commit_cptr = ffi.new('git_commit **')
-        refname_cstr = ffi.NULL
+        refname_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
         author_cptr = ffi.new('git_signature **')
         committer_cptr = ffi.new('git_signature **')
-        message_cstr = ffi.NULL
-        encoding_cstr = ffi.NULL
+        message_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
+        encoding_cstr: 'ArrayC[char]' | 'ffi.NULL_TYPE' = ffi.NULL
         tree_cptr = ffi.new('git_tree **')
 
         # Get commit as pointer to git_commit.
         if isinstance(commit, (str, Oid)):
-            commit = self[commit]
+            commit_object = self[commit]
+            commit_commit = commit_object.peel(Commit)
         elif isinstance(commit, Commit):
-            pass
+            commit_commit = commit
         elif commit is None:
             raise ValueError('the commit to amend cannot be None')
         else:
             raise TypeError('the commit to amend must be a Commit, str, or Oid')
-        commit = commit.peel(Commit)
-        ffi.buffer(commit_cptr)[:] = commit._pointer[:]
+        ffi.buffer(commit_cptr)[:] = commit_commit._pointer[:]
 
         # Get refname as C string.
         if isinstance(refname, Reference):
@@ -1553,9 +1712,11 @@ class BaseRepository(_Repository):
         # Get tree as pointer to git_tree.
         if tree is not None:
             if isinstance(tree, (str, Oid)):
-                tree = self[tree]
-            tree = tree.peel(Tree)
-            ffi.buffer(tree_cptr)[:] = tree._pointer[:]
+                tree_object = self[tree]
+            else:
+                tree_object = tree
+            tree_tree = tree_object.peel(Tree)
+            ffi.buffer(tree_cptr)[:] = tree_tree._pointer[:]
 
         # Amend the commit.
         err = C.git_commit_amend(
@@ -1572,11 +1733,16 @@ class BaseRepository(_Repository):
 
         return Oid(raw=bytes(ffi.buffer(coid)[:]))
 
+    def __ensure_tree(self, maybe_tree: str | Oid | Tree) -> Tree:
+        if isinstance(maybe_tree, Tree):
+            return maybe_tree
+        return self[maybe_tree].peel(Tree)
+
 
 class Repository(BaseRepository):
     def __init__(
         self,
-        path: typing.Optional[str] = None,
+        path: str | bytes | None | Path = None,
         flags: RepositoryOpenFlag = RepositoryOpenFlag.DEFAULT,
     ):
         """
@@ -1610,10 +1776,10 @@ class Repository(BaseRepository):
             super().__init__()
 
     @classmethod
-    def _from_c(cls, ptr, owned):
+    def _from_c(cls, ptr: 'GitRepositoryC', owned: bool) -> 'Repository':
         cptr = ffi.new('git_repository **')
         cptr[0] = ptr
         repo = cls.__new__(cls)
-        BaseRepository._from_c(repo, bytes(ffi.buffer(cptr)[:]), owned)
+        BaseRepository._from_c(repo, bytes(ffi.buffer(cptr)[:]), owned)  # type: ignore
         repo._common_init()
         return repo
